import bpy
from .logic_functions import*
from json import *
from functools import partial

component_prefix = "component_"
components_prop_name = "components"

def set_components(writer, values):
	for key, value in values.items():
		writer("{}{}".format(component_prefix, key), value)

def set_component(writer, query, value):
	writer("{}{}".format(component_prefix, query), value)
		
def get_components(getter):
	return {name[len(component_prefix):]: value for name, value in getter().items() if name.startswith(component_prefix)}

def get_component(getter, query):
	try:
		return getter("{}{}".format(component_prefix, query))
	except:
		return

def remove_component(remover, query):
	remover("{}{}".format(component_prefix, query))

def add_component(adder, query, val):
	adder("{]{}".format(component_prefix, query), val)
		
def createLogic(context, method):
	#Functions
	scene = context.scene
	current_object = context.object

	bpy.ops.scene.init_multiplayer()
	
	if method == "client":
		components = {
		"plugins.manager" : {}, 
		"terminal.client" : {
			"server_address" : scene.server_address,
			"server_port" : scene.server_port,
			"client_name" : scene.client_name, 
			"frequency" : scene.client_frequency,
			"timeout" : scene.time_out_client,
			}
		}
		
		
	elif method == "host":
		components = {
		"terminal.server" : {
			"player_limit" : scene.player_limit,
			"host_port" : scene.host_port,
			"timeout" : scene.time_out,
			}
		}	
		
	# Sensor frequency
	frequency = 0
	
	#############################################################################
	# Store properties that are only used by one script
	# Store by method -> Object ->properties
	# This only executes once, and saves the host object as the only key
	singleProperties = loads(scene.single_properties)
	multiLogic = loads(scene.multi_logic)
	
	singleProperties[method] = {current_object.name : []}
	multiLogic[method] = {current_object.name : []}

	#Add args property
	for component_name, component_data in components.items():
	
		getter = partial(getProperty, current_object)
		exists = get_component(getter, component_name) is not None

		writer = partial(modifyProperty if exists else createProperty, current_object)
		set_component(writer, component_name, str(component_data))
			
		# Create a single user property
		if not component_name in singleProperties[method][current_object.name]:
			singleProperties[method][current_object.name].append(component_name)
	
	# Create Logic Bricks
	sensor = getSensor(current_object, "multiplayer_tick")
	if not sensor:
		sensor = createSensor(current_object, "always", "multiplayer_tick", [True, False], frequency=frequency)  
	
	sensor.frequency = frequency
	
	if not sensor in multiLogic[method][current_object.name]:
		multiLogic[method][current_object.name].append(sensor.name)
				
	controller = getController(current_object, "multiplayer_wrapper")
	if not controller:
		controller = createController(current_object,  "multiplayer_wrapper", "python", "wrapper", sensor, "run")
	
	if not controller in multiLogic[method][current_object.name]:	
		multiLogic[method][current_object.name].append(controller.name)
	
	scene.single_properties = dumps(singleProperties)
	scene.multi_logic = dumps(multiLogic) 

def remover(obj, name):
	for i, c in enumerate(obj.game.properties):
		if c.name == name:
			bpy.ops.object.game_property_remove(index=i)
	
def removeLogic(context, method):

	current_object = context.object
	scene = context.scene  
	
	singleProperties = loads(scene.single_properties)
	multiLogic = loads(scene.multi_logic)
	
	objects = bpy.data.objects
	logic_holder = next(objects[k] for k in multiLogic[method].keys()) # Get the object that was first saved in the dict	

	scene.objects.active = logic_holder
	
	# Remove logic which we know is single user
	properties = singleProperties[method][logic_holder.name]
	logic_bricks = multiLogic[method][logic_holder.name]
	
	# Get all game component properties
	getter = lambda:{c.name: c.value for c in logic_holder.game.properties}
	components = get_components(getter)
	
	# Remove any registered properties
	for component_name, component_data in components.items():
		if component_name in properties:
			remove_component(partial(remover, logic_holder), component_name)	
			
	components = get_components(getter)	
	
	# And update the main list to reflect the change
	singleProperties[method].pop(logic_holder.name)
	
	# Remove logic and properties which may be multiuser
	if not components:
		controllers = [c for c in logic_holder.game.controllers]
		sensors = [s for s in logic_holder.game.sensors]
		
		total = controllers.copy()
		total.extend(sensors)
			
		logic_brick_items = [x for x in total if x.name in logic_bricks]	
		
		for brick in logic_brick_items:
			if hasattr(brick, "use_level"):
				# Is a sensor
				bpy.ops.logic.sensor_remove(sensor=brick.name, object=logic_holder.name)
			else:
				bpy.ops.logic.controller_remove(controller=brick.name, object = logic_holder.name)
					
	scene.objects.active = current_object
	scene.single_properties = dumps(singleProperties)
	scene.multi_logic = dumps(multiLogic) 
	
def createPluginLog(context, plugin):
	#Functions
	current_object = context.object
	scene = context.scene
	arg_list = scene.plugin_properties
	
	#############################################################################
	# Store properties that are only used by one script
	# Store by plugin -> Object ->properties
	# This only executes once, and saves the host current_object as the only key
	singleProperties = loads(current_object.single_properties)
	multiLogic = loads(current_object.multi_logic)
	
	singleProperties[plugin] = []
		
	multiLogic[plugin] = []
	#############################################################################
	
	component_normal = {k:v for k, v in loads(arg_list).items() if k == plugin}
	components = {"plugins."+k:{x:getattr(current_object, loads(current_object.converted_properties)[plugin][x]) for x, a in argdic.items()} for k, argdic in component_normal.items()}
	component_keys = [x for x in components.keys()]
	
	# Sensor frequency
	frequency = 0
	
	getter = partial(getProperty, current_object)

	#Add args property
	for component_name, component_data in components.items():
		exists = get_component(getter, component_name) is not None

		writer = partial(modifyProperty if exists else createProperty, current_object)
		set_component(writer, component_name, str(component_data))
			
		# Create a single user property
		if not component_name in singleProperties[plugin]:
			singleProperties[plugin].append(component_name)
		
	# Create Logic Bricks
	sensor = getSensor(current_object, "multiplayer_tick")
	if not sensor:
		sensor = createSensor(current_object, "always", "multiplayer_tick", [True, False], frequency=frequency)  
	
	sensor.frequency = frequency
	
	if not sensor.name in multiLogic[plugin]:
		multiLogic[plugin].append(sensor.name)
				
		
	controller = getController(current_object, "multiplayer_wrapper")
	if not controller:
		controller = createController(current_object,  "multiplayer_wrapper", "python", "wrapper", sensor, "run")
	
	if not controller.name in multiLogic[plugin]:	
		multiLogic[plugin].append(controller.name)
		
	current_object.single_properties = dumps(singleProperties)
	current_object.multi_logic = dumps(multiLogic)
	
def removePluginLog(context, plugin):
	current_object = context.object
	scene = context.scene  

	singleProperties = loads(current_object.single_properties)
	multiLogic = loads(current_object.multi_logic)
	
	# Remove logic which we know is single user
	properties = singleProperties[plugin]
	logic_bricks = multiLogic[plugin]
	
	# Get all game component properties
	getter = lambda:{c.name: c.value for c in current_object.game.properties}
	components = get_components(getter)
	
	# Remove any registered properties
	for component_name, component_data in components.items():
		if component_name in properties:
			remove_component(partial(remover, current_object), component_name)	
			
	components = get_components(getter)	
	
	# And update the main list to reflect the change
	singleProperties.pop(plugin)
	
	# Remove logic and properties which may be multiuser
	if not components:				
		controllers = [c for c in current_object.game.controllers]
		sensors = [s for s in current_object.game.sensors]
		
		total = []
		for c in controllers:
			total.append(c)
		for s in sensors:
			total.append(s)
			
		logic_brick_items = [x for x in total if x.name in logic_bricks]	
		
		for brick in logic_brick_items:
			if hasattr(brick, "use_level"):
				# Is a sensor
				bpy.ops.logic.sensor_remove(sensor=brick.name, object=current_object.name)
			else:
				bpy.ops.logic.controller_remove(controller=brick.name, object = current_object.name)

		multiLogic.pop(plugin)
	
	current_object.single_properties = dumps(singleProperties)
	current_object.multi_logic = dumps(multiLogic)